"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const NopCheck = () => true;
class Validator {
    constructor(shape, verbose = false) {
        this.shape = shape;
        this.verbose = verbose;
    }
    validate(input) {
        this.input = input;
        this.stack = [];
        const pass = this.validateObject(this.input, {
            type: 'object',
            shape: this.shape,
        });
        this.input = null;
        this.stack = [];
        return pass;
    }
    // Validation steps
    reject() {
        if (this.verbose) {
            console.log('Invalid input ', this.stack.join('.'));
        }
        return false;
    }
    validateObject(obj, spec) {
        if (spec.type !== 'object' || (!obj && !spec.optional))
            return this.reject();
        const { shape } = spec;
        for (const key in shape) {
            this.stack.push(key);
            if (!this.validateItem(obj[key], shape[key]))
                return this.reject();
            this.stack.pop();
        }
        return true;
    }
    validateItem(value, spec) {
        if (typeof spec === 'string') {
            return ((spec === 'any' && value !== undefined) ||
                (spec === 'forbidden' && value === undefined) ||
                typeof value === spec);
        }
        const { type, optional, check = NopCheck } = spec;
        const typeCheck = (optional && (value === undefined || value === null)) ||
            (type === 'boolean' && typeof value === 'boolean') ||
            (type === 'string' && this.validateString(value, spec)) ||
            (type === 'number' && this.validateNumber(value, spec)) ||
            (type === 'object' && this.validateObject(value, spec)) ||
            (type === 'array' && this.validateArray(value, spec)) ||
            (type === 'enum' && this.validateEnum(value, spec));
        return typeCheck && check(value, this.input);
    }
    validateString(value, spec) {
        return (spec.type === 'string' &&
            (!spec.minLength || value.length >= spec.minLength) &&
            (!spec.maxLength || value.length <= spec.maxLength) &&
            (!spec.pattern || spec.pattern.test(value)));
    }
    validateNumber(value, spec) {
        return (spec.type === 'number' &&
            (!spec.min || value >= spec.min) &&
            (!spec.max || value <= spec.max) &&
            (!spec.integer || Number.isInteger(value)));
    }
    validateArray(arr, spec) {
        return (spec.type === 'array' &&
            (!spec.minLength || arr.length >= spec.minLength) &&
            (!spec.maxLength || arr.length <= spec.maxLength) &&
            arr.every(item => this.validateItem(item, spec.of)));
    }
    validateEnum(value, spec) {
        return spec.type === 'enum' && spec.of.includes(value);
    }
}
exports.Validator = Validator;
