"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const pg_1 = require("pg");
const pg_cursor_1 = __importDefault(require("pg-cursor"));
class Sql {
    constructor(db = Sql.pool) {
        this.db = db;
    }
    static createPool(config) {
        this.pool = new pg_1.Pool(config.pool);
    }
    async one(strings, ...merges) {
        const { rows } = await this.tagQuery(strings, merges);
        return rows[0];
    }
    async q(strings, ...merges) {
        const { rows } = await this.tagQuery(strings, merges);
        return rows;
    }
    async cur(strings, ...merges) {
        const { text, values } = buildQuery(strings, merges);
        const pgCursor = this.db.query(new pg_cursor_1.default(text, values));
        return {
            close: () => pgCursor.close(),
            next: count => new Promise((resolve, reject) => pgCursor.read(count, (err, rows) => err ? reject(err) : resolve(rows))),
        };
    }
    async raw(query, values = []) {
        const { rows } = await this.exec(query, values);
        return rows;
    }
    // SINGLE-ROW OPERATIONS
    async get(table, value, matchBy = 'id') {
        const sql = `SELECT * FROM ${table} WHERE ${matchBy} = $1`;
        const { rows } = await this.exec(sql, [value]);
        return rows[0];
    }
    async value(sql, values, name = 'result') {
        const { rows } = await this.exec(sql, values);
        return rows[0][name];
    }
    async insert(table, row, returning = '*') {
        const columns = Object.keys(row).join(',');
        const values = Object.values(row);
        const placeholders = values.map((_, i) => `$${i + 1}`);
        const sql = `
      INSERT INTO ${table} (${columns})
      VALUES (${placeholders})
      RETURNING ${returning}
    `;
        const { rows } = await this.exec(sql, values);
        return rows[0];
    }
    async update(table, row, matchBy = 'id') {
        const matchValue = row[matchBy];
        const changes = { ...row };
        delete changes[matchBy];
        const columns = Object.keys(changes);
        const values = [];
        const pairs = [];
        for (const col of columns) {
            values.push(changes[col]);
            pairs.push(`${col} = $${values.length}`);
        }
        values.push(matchValue);
        const sql = `
      UPDATE ${table}
      SET ${pairs.join(',')}
      WHERE ${matchBy} = ${'$' + values.length}
    `;
        const result = await this.exec(sql, values);
        return result.rowCount === 1;
    }
    async delete(table, row, matchBy = 'id') {
        const sql = `DELETE FROM ${table} WHERE ${matchBy} = $1`;
        const values = [row[matchBy]];
        const result = await this.exec(sql, values);
        return result.rowCount === 1;
    }
    // MISCELLANEOUS
    async transaction(block) {
        if (this.db !== Sql.pool)
            throw Error('Already in transaction');
        const client = await Sql.pool.connect();
        await client.query('BEGIN');
        const tx = new Sql(client);
        let result = true;
        try {
            result = await block(tx);
        }
        catch (e) {
            result = false;
        }
        await client.query(result !== false ? 'COMMIT' : 'ROLLBACK');
        client.release();
        return result;
    }
    async arrayTransaction(items, operation) {
        if (this.db !== Sql.pool)
            throw Error('Already in transaction');
        const client = await Sql.pool.connect();
        await client.query('BEGIN');
        const tx = new Sql(client);
        let result = null;
        try {
            result = await Promise.all(items.map((item, i) => operation(tx, item, i)));
        }
        catch (e) {
            result = null;
        }
        await client.query(result !== null ? 'COMMIT' : 'ROLLBACK');
        client.release();
        return result;
    }
    // INTERNAL
    tagQuery(strings, merges) {
        const { text, values } = buildQuery(strings, merges);
        return this.exec(text, values);
    }
    async exec(text, values = []) {
        try {
            const result = await this.db.query({ text, values });
            return result;
        }
        catch (e) {
            const logMessage = [
                `SQL query error: ${e.message}`,
                text,
                JSON.stringify(values),
            ].join('\n');
            throw Error(logMessage);
        }
    }
}
exports.Sql = Sql;
function buildQuery(strings, merges) {
    const query = [];
    const values = [];
    for (let i = 0; i < strings.length; i++) {
        query.push(strings[i]);
        if (i >= merges.length)
            continue;
        const merge = merges[i];
        if (!Array.isArray(merge)) {
            query.push(merge);
        }
        else if (merge.length === 1) {
            values.push(merge[0]);
            query.push('$' + values.length);
        }
        else {
            values.push(`{${merge.join(',')}}`);
            query.push(`ANY($${values.length})`);
        }
    }
    return { text: query.join(''), values };
}
exports.buildQuery = buildQuery;
