"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const pg_1 = require("pg");
const SqlMacros_1 = require("./SqlMacros");
class SqlInstaller {
    constructor(config, versionsDir, log) {
        this.config = config;
        this.versionsDir = versionsDir;
        this.log = log;
        this.verbose = process.env.NODE_ENV === 'dev';
    }
    async run() {
        this.log.info('Start');
        try {
            await this.connect();
            const upgrades = await this.getUpgrades();
            for (const version of upgrades) {
                await this.doUpgrade(version);
            }
            this.log.info('Complete');
            return true;
        }
        catch (e) {
            this.log.error('Failed');
            this.log.error(e);
            if (this.db) {
                await this.db.query('ROLLBACK').catch();
                await this.db.end().catch();
            }
            return false;
        }
    }
    async connect() {
        const db = new pg_1.Client(this.config.connection);
        await db.connect();
        this.db = db;
    }
    async getUpgrades() {
        const versions = fs_1.default.readdirSync(this.versionsDir).sort((one, other) => {
            const [m1, s1, p1] = one.split('.').map(Number);
            const [m2, s2, p2] = other.split('.').map(Number);
            return m1 - m2 || s1 - s2 || p1 - p2;
        });
        const latestRow = await this.db.query(`
      SELECT semver
      FROM ${this.config.versionsTable}
      ORDER BY sequence DESC
      LIMIT 1
    `);
        const current = latestRow.rows[0].semver;
        return versions.slice(versions.indexOf(current) + 1);
    }
    async doUpgrade(version) {
        this.log.info('Upgrading to', version);
        await this.db.query('BEGIN');
        const versionDir = path_1.default.join(this.versionsDir, version);
        const sqlFiles = fs_1.default
            .readdirSync(versionDir)
            .filter((name) => name.endsWith('.sql'))
            .sort((one, other) => (one === 'core.sql' && -1) ||
            (other === 'core.sql' && 1) ||
            one.localeCompare(other));
        for (const fileName of sqlFiles) {
            await this.runSqlFile(versionDir, fileName);
            await this.runPostScript(versionDir, fileName);
        }
        await this.updateVersionLog(version, versionDir);
        await this.db.query('COMMIT');
    }
    async runSqlFile(versionDir, fileName) {
        const sqlFile = path_1.default.join(versionDir, fileName);
        const commands = fs_1.default
            .readFileSync(sqlFile, { encoding: 'utf8' })
            .split(';')
            .filter((line) => line.trim())
            .map(SqlMacros_1.applyMacros);
        for (const command of commands) {
            if (this.verbose) {
                this.log.debug(command);
            }
            await this.db.query(command);
            this.log.info('OK');
        }
    }
    async runPostScript(versionDir, fileName) {
        const postFile = path_1.default.join(versionDir, fileName + '.js');
        if (!fs_1.default.existsSync(postFile))
            return;
        const postScript = require(postFile);
        await postScript(this.db);
    }
    async updateVersionLog(version, versionDir) {
        const notesFile = path_1.default.join(versionDir, 'notes.json');
        const notes = fs_1.default.existsSync(notesFile)
            ? JSON.parse(fs_1.default.readFileSync(notesFile, { encoding: 'utf8' }))
            : {};
        await this.db.query(`
      INSERT INTO ${this.config.versionsTable} (semver, notes)
      VALUES ($1, $2)
      `, [version, notes.descr || '']);
    }
}
exports.SqlInstaller = SqlInstaller;
