import { Complex, Primitive, Shape } from './types';

const NopCheck = () => true;

export class Validator {
  private input: any;
  private stack: string[];

  constructor(
    private readonly shape: Shape,
    private readonly verbose: boolean = false,
  ) {}

  validate(input: any): boolean {
    this.input = input;
    this.stack = [];

    const pass = this.validateObject(this.input, {
      type: 'object',
      shape: this.shape,
    });

    this.input = null;
    this.stack = [];
    return pass;
  }

  // Validation steps

  private reject(): false {
    if (this.verbose) {
      console.log('Invalid input ', this.stack.join('.'));
    }

    return false;
  }

  private validateObject(obj: object, spec: Complex): boolean {
    if (spec.type !== 'object' || (!obj && !spec.optional))
      return this.reject();

    const { shape } = spec;
    for (const key in shape) {
      this.stack.push(key);
      if (!this.validateItem(obj[key], shape[key])) return this.reject();
      this.stack.pop();
    }
    return true;
  }

  private validateItem(value: any, spec: Primitive | Complex) {
    if (typeof spec === 'string') {
      return (
        (spec === 'any' && value !== undefined) ||
        (spec === 'forbidden' && value === undefined) ||
        typeof value === spec
      );
    }

    const { type, optional, check = NopCheck } = spec as Complex;
    const typeCheck =
      (optional && (value === undefined || value === null)) ||
      (type === 'boolean' && typeof value === 'boolean') ||
      (type === 'string' && this.validateString(value, spec as Complex)) ||
      (type === 'number' && this.validateNumber(value, spec as Complex)) ||
      (type === 'object' && this.validateObject(value, spec as Complex)) ||
      (type === 'array' && this.validateArray(value, spec as Complex)) ||
      (type === 'enum' && this.validateEnum(value, spec as Complex));

    return typeCheck && check(value, this.input);
  }

  private validateString(value: string, spec: Complex): boolean {
    return (
      spec.type === 'string' &&
      (!spec.minLength || value.length >= spec.minLength) &&
      (!spec.maxLength || value.length <= spec.maxLength) &&
      (!spec.pattern || spec.pattern.test(value))
    );
  }

  private validateNumber(value: number, spec: Complex): boolean {
    return (
      spec.type === 'number' &&
      (!spec.min || value >= spec.min) &&
      (!spec.max || value <= spec.max) &&
      (!spec.integer || Number.isInteger(value))
    );
  }

  private validateArray(arr: Array<any>, spec: Complex): boolean {
    return (
      spec.type === 'array' &&
      (!spec.minLength || arr.length >= spec.minLength) &&
      (!spec.maxLength || arr.length <= spec.maxLength) &&
      arr.every(item => this.validateItem(item, spec.of))
    );
  }

  private validateEnum(value: string, spec: Complex): boolean {
    return spec.type === 'enum' && spec.of.includes(value);
  }
}
