import { Pool, QueryConfig, QueryResult, PoolConfig } from 'pg';
import Cursor from 'pg-cursor';
import { SqlClient, SqlCursor } from '../exports/SqlClient';

interface Connection {
  query<T>(query: string, values?: any[]): Promise<T>;
  query<T>(config: QueryConfig): Promise<QueryResult<T>>;
  release?(): void;
}

export interface SqlConfig {
  pool: PoolConfig;
}

export class Sql implements SqlClient {
  private static pool: Pool;

  static createPool(config: SqlConfig) {
    this.pool = new Pool(config.pool);
  }

  constructor(private db: Connection = Sql.pool) {}

  async one<T = any>(
    strings: TemplateStringsArray,
    ...merges: any[]
  ): Promise<T> {
    const { rows } = await this.tagQuery<T>(strings, merges);
    return rows[0];
  }

  async q<T = any>(
    strings: TemplateStringsArray,
    ...merges: any[]
  ): Promise<T[]> {
    const { rows } = await this.tagQuery<T>(strings, merges);
    return rows;
  }

  async cur<T = any>(
    strings: TemplateStringsArray,
    ...merges: any[]
  ): Promise<SqlCursor<T>> {
    const { text, values } = buildQuery(strings, merges);
    const pgCursor = this.db.query(new Cursor(text, values)) as any;
    return {
      close: () => pgCursor.close(),
      next: count =>
        new Promise((resolve, reject) =>
          pgCursor.read(count, (err: Error, rows: T[]) =>
            err ? reject(err) : resolve(rows),
          ),
        ),
    };
  }

  async raw<T = any>(query: string, values: any[] = []): Promise<T[]> {
    const { rows } = await this.exec<T>(query, values);
    return rows;
  }

  // SINGLE-ROW OPERATIONS

  async get<T = any>(
    table: string,
    value: any,
    matchBy: string = 'id',
  ): Promise<T> {
    const sql = `SELECT * FROM ${table} WHERE ${matchBy} = $1`;
    const { rows } = await this.exec(sql, [value]);
    return rows[0] as T;
  }

  async value<T = any>(
    sql: string,
    values: any[],
    name: string = 'result',
  ): Promise<T> {
    const { rows } = await this.exec(sql, values);
    return rows[0][name] as T;
  }

  async insert<T = any>(
    table: string,
    row: Partial<T>,
    returning: string = '*',
  ): Promise<T> {
    const columns = Object.keys(row).join(',');
    const values = Object.values(row);
    const placeholders = values.map((_, i) => `$${i + 1}`);

    const sql = `
      INSERT INTO ${table} (${columns})
      VALUES (${placeholders})
      RETURNING ${returning}
    `;
    const { rows } = await this.exec(sql, values);
    return rows[0] as T;
  }

  async update<T = any>(
    table: string,
    row: Partial<T>,
    matchBy: string = 'id',
  ): Promise<boolean> {
    const matchValue = row[matchBy];

    const changes = { ...row };
    delete changes[matchBy];
    const columns = Object.keys(changes);
    const values = [];
    const pairs = [];
    for (const col of columns) {
      values.push(changes[col]);
      pairs.push(`${col} = $${values.length}`);
    }
    values.push(matchValue);

    const sql = `
      UPDATE ${table}
      SET ${pairs.join(',')}
      WHERE ${matchBy} = ${'$' + values.length}
    `;
    const result = await this.exec(sql, values);
    return result.rowCount === 1;
  }

  async delete<T = any>(
    table: string,
    row: Partial<T>,
    matchBy: string = 'id',
  ): Promise<boolean> {
    const sql = `DELETE FROM ${table} WHERE ${matchBy} = $1`;
    const values = [row[matchBy]];
    const result = await this.exec(sql, values);
    return result.rowCount === 1;
  }

  // MISCELLANEOUS

  async transaction<R = any>(
    block: (tx: Sql) => Promise<R | boolean>,
  ): Promise<R> {
    if (this.db !== Sql.pool) throw Error('Already in transaction');

    const client = await Sql.pool.connect();
    await client.query('BEGIN');

    const tx = new Sql(client);
    let result: R | boolean = true;
    try {
      result = await block(tx);
    } catch (e) {
      result = false;
    }

    await client.query(result !== false ? 'COMMIT' : 'ROLLBACK');
    client.release();

    return result as R;
  }

  async arrayTransaction<T = any, R = any>(
    items: T[],
    operation: (tx: Sql, item: T, i: number) => Promise<R>,
  ): Promise<R[]> {
    if (this.db !== Sql.pool) throw Error('Already in transaction');

    const client = await Sql.pool.connect();
    await client.query('BEGIN');

    const tx = new Sql(client);
    let result: R[] = null;
    try {
      result = await Promise.all<R>(
        items.map((item, i) => operation(tx, item, i)),
      );
    } catch (e) {
      result = null;
    }

    await client.query(result !== null ? 'COMMIT' : 'ROLLBACK');
    client.release();

    return result;
  }

  // INTERNAL

  private tagQuery<T>(
    strings: TemplateStringsArray,
    merges: any[],
  ): Promise<QueryResult<T>> {
    const { text, values } = buildQuery(strings, merges);
    return this.exec<T>(text, values);
  }

  private async exec<T = any>(text: string, values: any[] = []) {
    try {
      const result = await this.db.query<T>({ text, values });
      return result;
    } catch (e) {
      const logMessage = [
        `SQL query error: ${e.message}`,
        text,
        JSON.stringify(values),
      ].join('\n');
      throw Error(logMessage);
    }
  }
}

export function buildQuery(
  strings: TemplateStringsArray,
  merges: any[],
): QueryConfig {
  const query = [];
  const values = [];
  for (let i = 0; i < strings.length; i++) {
    query.push(strings[i]);
    if (i >= merges.length) continue;

    const merge = merges[i];
    if (!Array.isArray(merge)) {
      query.push(merge);
    } else if (merge.length === 1) {
      values.push(merge[0]);
      query.push('$' + values.length);
    } else {
      values.push(`{${merge.join(',')}}`);
      query.push(`ANY($${values.length})`);
    }
  }

  return { text: query.join(''), values };
}
