import { Request, Response } from 'express';
import * as jwt from 'jsonwebtoken';

export interface UserConfig {
  cookieName: string;
  cookie: {
    path: string;
    domain: string;
    maxAge: number;
  };
  jwtSecret: string;
  hardAge: number;
  payloadFields: string[];
}

interface TokenPayload {
  hard: number;
  user: any;
}

export class User {
  private state: 'guest' | 'hard' | 'soft';
  private data: any;

  constructor(
    private readonly isProd: boolean,
    private readonly config: UserConfig,
    private readonly req: Request,
    private readonly res: Response,
  ) {}

  authenticate() {
    const { cookieName } = this.config;
    const { [cookieName]: token } = this.req.cookies;
    const payload = this.verify(token);
    if (!payload) {
      this.logout();
      return;
    }

    const { user, hard } = payload;
    this.data = user;
    this.state = Date.now() <= hard ? 'hard' : 'soft';
    this.setCookie(token);
  }

  login(data: any, hard = false) {
    this.data = data;
    this.state = hard ? 'hard' : 'soft';

    const { hardAge, jwtSecret } = this.config;
    const payload = {
      hard: hard ? Date.now() + hardAge : 0,
      user: {},
    };
    for (const field of this.config.payloadFields) {
      payload.user[field] = data[field];
    }

    const token = jwt.sign(payload, jwtSecret);
    this.setCookie(token);
  }

  logout() {
    this.state = 'guest';
    this.data = null;
    this.res.clearCookie(this.config.cookieName);
  }

  isGuest(): boolean {
    return this.state === 'guest';
  }

  isHard(): boolean {
    return this.state === 'hard';
  }

  get(): any {
    return this.data;
  }

  getRemoteIp() {
    return this.req.get('X-Client-IP');
  }

  // INTERNAL

  private verify(token: string): TokenPayload {
    if (!token) return null;

    try {
      const { jwtSecret } = this.config;
      const payload = jwt.verify(token, jwtSecret) as TokenPayload;
      return payload;
    } catch (e) {
      return null;
    }
  }

  private setCookie(token: string) {
    this.res.cookie(this.config.cookieName, token, {
      ...this.config.cookie,

      httpOnly: this.isProd,
      secure: this.isProd,
      sameSite: 'lax',
    });
  }
}
