import fs from 'fs';
import path from 'path';
import { Client, ClientConfig } from 'pg';

import { Log } from '../Log';
import { applyMacros } from './SqlMacros';

type PostScript = (db: Client) => Promise<void>;

export interface SqlInstallerConfig {
  connection: ClientConfig;
  versionsTable: string;
}

export class SqlInstaller {
  private readonly verbose = process.env.NODE_ENV === 'dev';
  private db: Client;

  constructor(
    private readonly config: SqlInstallerConfig,
    private readonly versionsDir: string,
    private readonly log: Log,
  ) {}

  async run(): Promise<boolean> {
    this.log.info('Start');

    try {
      await this.connect();

      const upgrades = await this.getUpgrades();
      for (const version of upgrades) {
        await this.doUpgrade(version);
      }

      this.log.info('Complete');
      return true;
    } catch (e) {
      this.log.error('Failed');
      this.log.error(e);

      if (this.db) {
        await this.db.query('ROLLBACK').catch();
        await this.db.end().catch();
      }
      return false;
    }
  }

  private async connect() {
    const db = new Client(this.config.connection);

    await db.connect();
    this.db = db;
  }

  private async getUpgrades(): Promise<string[]> {
    const versions = fs.readdirSync(this.versionsDir).sort((one, other) => {
      const [m1, s1, p1] = one.split('.').map(Number);
      const [m2, s2, p2] = other.split('.').map(Number);
      return m1 - m2 || s1 - s2 || p1 - p2;
    });

    const latestRow = await this.db.query(`
      SELECT semver
      FROM ${this.config.versionsTable}
      ORDER BY sequence DESC
      LIMIT 1
    `);
    const current = latestRow.rows[0].semver;

    return versions.slice(versions.indexOf(current) + 1);
  }

  private async doUpgrade(version: string) {
    this.log.info('Upgrading to', version);

    await this.db.query('BEGIN');

    const versionDir = path.join(this.versionsDir, version);
    const sqlFiles = fs
      .readdirSync(versionDir)
      .filter((name) => name.endsWith('.sql'))
      .sort(
        (one, other) =>
          (one === 'core.sql' && -1) ||
          (other === 'core.sql' && 1) ||
          one.localeCompare(other),
      );
    for (const fileName of sqlFiles) {
      await this.runSqlFile(versionDir, fileName);
      await this.runPostScript(versionDir, fileName);
    }

    await this.updateVersionLog(version, versionDir);
    await this.db.query('COMMIT');
  }

  private async runSqlFile(versionDir: string, fileName: string) {
    const sqlFile = path.join(versionDir, fileName);
    const commands = fs
      .readFileSync(sqlFile, { encoding: 'utf8' })
      .split(';')
      .filter((line) => line.trim())
      .map(applyMacros);
    for (const command of commands) {
      if (this.verbose) {
        this.log.debug(command);
      }
      await this.db.query(command);
      this.log.info('OK');
    }
  }

  private async runPostScript(versionDir: string, fileName: string) {
    const postFile = path.join(versionDir, fileName + '.js');
    if (!fs.existsSync(postFile)) return;

    const postScript = require(postFile) as PostScript;
    await postScript(this.db);
  }

  private async updateVersionLog(version: string, versionDir: string) {
    const notesFile = path.join(versionDir, 'notes.json');
    const notes = fs.existsSync(notesFile)
      ? JSON.parse(fs.readFileSync(notesFile, { encoding: 'utf8' }))
      : {};
    await this.db.query(
      `
      INSERT INTO ${this.config.versionsTable} (semver, notes)
      VALUES ($1, $2)
      `,
      [version, notes.descr || ''],
    );
  }
}
